%%
%
% Boosting methods
% 1. Boosted Random Fern
% 2. AdaBoost
%
ccc
rng(1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 1. Generate training data
nr_train_pos = 200; nr_train_neg = 200; nr_train = nr_train_pos + nr_train_neg;
nr_test_pos = 500; nr_test_neg = 500; nr_test = nr_test_pos + nr_test_neg;
tr_data = gen_spiral(nr_train_pos, nr_train_neg);
te_data = gen_spiral(nr_test_pos, nr_test_neg);

% 2. Plot
if 1
    figure(1);
    subplot(121); hold on;
    pos_idx = find(tr_data(:, 3) == 1);
    neg_idx = find(tr_data(:, 3) == -1);
    plot(tr_data(pos_idx, 1), tr_data(pos_idx, 2), 'ro');
    plot(tr_data(neg_idx, 1), tr_data(neg_idx, 2), 'bx');
    axis equal; grid on; title('Training Data');
    subplot(122); hold on;
    pos_idx = find(te_data(:, 3) == 1);
    neg_idx = find(te_data(:, 3) == -1);
    plot(te_data(pos_idx, 1), te_data(pos_idx, 2), 'ro');
    plot(te_data(neg_idx, 1), te_data(neg_idx, 2), 'bx');
    axis equal; grid on; title('Test Data');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% A. Boosted Random Fern (BRF)
% Train BRF
nr_fern = 3000;
nr_feat = 7;
ferns   = init_ferns(nr_fern, nr_feat);
nr_weak_brf = 50;    % Weak learner 
tic
brf     = train_brf(ferns, tr_data, nr_weak_brf);
brf_train_toc = toc;
% Test BRF
brf_train_pred    = testonly_rf(brf, tr_data);
tic;
brf_test_pred     = testonly_rf(brf, te_data);
brf_test_toc = toc;
brf_train_correct = sign(brf_train_pred) == tr_data(:, 3);
brf_test_correct  = sign(brf_test_pred) == te_data(:, 3);
brf_train_acc     = mean(brf_train_correct);
brf_test_acc      = mean(brf_test_correct);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% B. AdaBoost
% Train AdaBoost
nr_weak_ada = 350;    % Weak learner 
tr_set = tr_data(:, 1:2); tr_labels = zeros(nr_train, 1);
pidx = find(tr_data(:, 3) == -1); nidx = find(tr_data(:, 3) == 1);
tr_labels(pidx) = 1; tr_labels(nidx) = 2;
te_set = te_data(:, 1:2); te_labels = zeros(nr_test, 1);
pidx2 = find(te_data(:, 3) == -1); nidx2 = find(te_data(:, 3) == 1);
te_labels(pidx2) = 1; te_labels(nidx2) = 2;
tic;
adaboost = train_adaboost(@threshold_tr, @threshold_te, tr_set, tr_labels, nr_weak_ada);
adaboost_train_toc = toc;
% Test AdaBoost
preds_tr = test_adaboost(adaboost, @threshold_te, tr_set);
tic;
preds_te = test_adaboost(adaboost, @threshold_te, te_set);
adaboost_test_toc  = toc;
adaboost_train_acc = mean(preds_tr==tr_labels);
adaboost_test_acc  = mean(preds_te==te_labels);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% C. Show
fprintf('======== Boosted Random Fern (%d-depth %d trees) ========\n', nr_feat, nr_weak_brf);
fprintf('BFR train: %.3f sec \n', brf_train_toc);
fprintf('BFR test: %.3f sec \n', brf_test_toc);
fprintf('BRF Train Accuracy: %.1f%% BRF Test Accuracy: %.1f%% \n', brf_train_acc*100, brf_test_acc*100);

fprintf('======== AdaBoost (%d decision stumps) ========\n', nr_weak_ada);
fprintf('AdaBoost train: %.3f sec \n', adaboost_train_toc);
fprintf('AdaBoost test: %.3f sec \n', adaboost_test_toc);
fprintf('AdaBoost Train Accuracy: %.1f%% AdaBoost Test Accuracy: %.1f%%  \n', adaboost_train_acc*100, adaboost_test_acc*100);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Do some plotting
xgrid = linspace(0, 1, 50);
ygrid = linspace(0, 1, 50);
[mesh_X, mesh_Y] = meshgrid(xgrid, ygrid);
points = [mesh_X(:) mesh_Y(:)];
nr_points = length(points);
% Boosted Random Fern
brf_grid = testonly_rf(brf, points);
brf_grid = brf_grid - min(brf_grid);
brf_grid = brf_grid / max(brf_grid);
brf_grid_rs = reshape(brf_grid, 50, 50);
% AdaBoost
[preds_te, L_te] = test_adaboost(adaboost, @threshold_te, points);
L_temp = -L_te(:, 1) + L_te(:, 2);
L_temp = L_temp - min(L_temp);
L_temp = L_temp / max(L_temp);
ada_grid_rs = reshape(L_temp, 50, 50);

figure(2); clf; colormap jet
imagesc(ygrid, xgrid, brf_grid_rs); hold on;
pos_idx = find(tr_data(:, 3) == 1);
neg_idx = find(tr_data(:, 3) == -1);
plot(tr_data(pos_idx, 1), tr_data(pos_idx, 2), 'ko', 'MarkerFaceColor', 'r');
plot(tr_data(neg_idx, 1), tr_data(neg_idx, 2), 'k^', 'MarkerFaceColor', 'b');
axis equal; axis off; grid on; title('Training Data');
colorbar; title('Boosted Random Fern');


figure(3); clf; colormap jet
imagesc(ygrid, xgrid, ada_grid_rs); hold on;
pos_idx = find(tr_data(:, 3) == 1);
neg_idx = find(tr_data(:, 3) == -1);
plot(tr_data(pos_idx, 1), tr_data(pos_idx, 2), 'ko', 'MarkerFaceColor', 'r');
plot(tr_data(neg_idx, 1), tr_data(neg_idx, 2), 'k^', 'MarkerFaceColor', 'b');
axis equal; axis off; grid on; title('Training Data');
colorbar; title('AdaBoost');

%%
